/*
 * File : dlInfoXfer.C
 *
 * Description : This file contains the method definitions for the
 * dlInfoXfer class.
 */

#include "dlDisconnected.h"
#include "frame.h"
#include "dataPacket.h"
#include "dllayer.h"
#include "dlEvents.h"
#include "phEvents.h"
#include "dlInfoXfer.h"
#include "dlDiscPending.h"
#include "dlInSetupPending.h"
#include "dlOutSetupPending.h"
#include "dlFrameReject.h"
#include "dlReceiverBusy.h"
#include "dlReject.h"

// the receiverBusy variable tells us whether we should
// should go into busy state or not. It is set in the nwlayer/user
// interface. The user can toggle the value of this variable
// from TRUE to FALSE and vice-versa.

extern Boolean receiverBusy;

/*
 * Function : dlInfoXfer()
 *
 * Description : This is the constructor. Invokes base class
 * constructor. If 'reset' is TRUE, then the state variables
 * of the data-link layer are reset. 'reset' is FALSE whenever
 * the state is being changed from dlReceiverBusy to dlInfoXfer,
 * or from dlReject to dlInfoXfer (since they are essentially
 * special conditions of the information transfer phase). 
 * 'reset' is TRUE whenever dlInfoXfer state is entered after
 * a link setup, or a link reset.
 */

dlInfoXfer::dlInfoXfer(dllayer *dl, Boolean reset) : dlStateMachine(dl) 
{
	if ( reset == TRUE )
		_dl->ResetStateVariables();
}

/*
 * Function : PrintState()
 *
 * Description : Prints the state.
 */

void 
dlInfoXfer::PrintState()
{
	TRACE(dlSmModule,1,("In dlInfoXfer\n"));
}

/*
 * Function : DlDisconnectReq()
 *
 * Description : Handles a DL_DisconnectReq primitive invoked
 * by the network layer. Transmits a DISC, and changes
 * state to dlDiscPending.
 */

dlStateMachine *
dlInfoXfer::DlDisconnectReq(int &errorCode) 
{
	TRACE(dlSmModule,1,("Entering DlDisconnectReq\n"));

	// Transmit a DISC, P=1, and start timer T1

	_dl->XmitFrame(errorCode,TRUE,DISC,1);
	if ( errorCode == codeSuccessful ) {		
		_dl->StartT1Timer(FALSE);					
		CHANGE_STATE(dlDiscPending);
	}
	else 
		return this;
}

/*
 * Function : HandleSabm()
 *
 * Description : Processes a SABM that was received.
 * This is request for link reset. Invoke DL_ResetInd
 * on the network layer, and change state to 
 * dlInSetupPending. Note setup type is externalReset. 
 */

dlStateMachine *
dlInfoXfer::HandleSabm(int &errorCode) 
{
	TRACE(dlSmModule,1,("In HandleSabm\n"));

	_dl->_nPlus1->Command(DL_ResetInd,errorCode,peerReset);
	_dl->_setupType = externalReset;
	CHANGE_STATE(dlInSetupPending);
}

/*
 * Function : HandleDisc()
 *
 * Description : Processes a DISC that was received.
 * This is request for link disconnect. Invoke DL_DisconnectInd
 * on the network layer, and change state to dlDisconnected. 
 */

dlStateMachine *
dlInfoXfer::HandleDisc(int &errorCode) 
{
	TRACE(dlSmModule,1,("In HandleDisc\n"));

	// First, transmit a UA
	_dl->XmitFrame(errorCode,FALSE,UA,_dl->_lastFrameRecd->GetPf());

	// Notify the network layer
	_dl->_nPlus1->Command(DL_DisconnectInd,errorCode,peerDisconnect);

	// Change state..
	CHANGE_STATE(dlDisconnected);
}

/*
 * Function : HandleUA()
 *
 * Description :  Processes a UA that was received.
 * Call LinkReset().
 */

dlStateMachine *
dlInfoXfer::HandleUa(int &errorCode) 
{
	return (LinkReset(errorCode));
}

/*
 * Function : HandleDM()
 *
 * Description :  Processes a DM that was received.
 * Call LinkReset().
 */

dlStateMachine *
dlInfoXfer::HandleDm(int &errorCode) 
{
	return (LinkReset(errorCode));
}

/*
 * Function : HandleFrmr()
 *
 * Description : Processes a FRMR that was received.
 *	Calls LinkReset().
 */

dlStateMachine *
dlInfoXfer::HandleFrmr(int &errorCode) 
{
	dataPacket *data = _dl->_lastFrameRecd->GetDataPacket();
	TRACE(dlSmModule,1,("Received FRMR, INFO LENGTH : %d\n",data->Length())); 
	data->Print(); // For Debugging..	
	return (LinkReset(errorCode));
}

/*
 * Function : HandleUnsolFBit()
 *
 * Description : Processes a response frame received
 *	with an unsolicited F bit set to 1.
 * Calls LinkReset().
 */

dlStateMachine *
dlInfoXfer::HandleUnsolFBit(int &errorCode)
{
	TRACE(dlSmModule,1,("Entering UnsolFBit\n"));
	return (LinkReset(errorCode));
}

/*
 * Function : LinkReset()
 *
 * Description : Transmits a SABM, starts T1 timer
 * and changes state to dlOutSetupPending.
 */

dlStateMachine *
dlInfoXfer::LinkReset(int &errorCode)
{
	_dl->XmitFrame(errorCode,TRUE,SABM,1);
	if ( errorCode == codeSuccessful ) {		
		_dl->StartT1Timer(FALSE);					
		_dl->_setupType = internalReset;
		CHANGE_STATE(dlOutSetupPending);
	}
	else 
		return this;
}

/*
 * Function : DlDataReq()
 *
 * Description : Handles a DL_DataReq primitive invoked
 *	by the network layer. The dataPacket passed down
 * by the network layer as part of the primitive 
 * invocation has already been converted into a frame
 * object and queued in the window object (_dl->_win).
 * I-frames are not transmitted if the data-link is 
 * in a checkpointing cycle, or if the remote DTE is
 * in a BUSY state.
 */

dlStateMachine *
dlInfoXfer::DlDataReq(int &errorCode)
{
	TRACE(dlSmModule,1,("Entering DlDataReq\n"));

	// First check if remote DTE is busy
	// Next check if we are in checkpointing..
	// All clear - transmit frames.

	if ( _dl->_remoteBusy == TRUE ) {
		TRACE(dlSmModule,1,("Remote BUSY\n"));
		// Remote DTE is busy
		// As far the network layer is concerned 
		// accept the data successfully for later transmission
		errorCode = codeSuccessful; 
	}
	else if ( _dl->_t1Timer != 0 ) {
		TRACE(dlSmModule,1,("In CheckPointing cycle\n"));
		// As far the network layer is concerned 
		// accept the data successfully for later transmission
		errorCode = codeSuccessful; 
	}
	else {
		// Transmit the I-frames, start timer etc.
		Transmit(FALSE);
	}
	return this;
}

/*
 * Function : DlResetReq()
 *
 * Description : Handles a DL_ResetReq primitive invoked by
 * the network layer. Transmits a SABM, starts T1 timer,
 * and changes state to dlOutSetupPending.
 */

dlStateMachine *
dlInfoXfer::DlResetReq(int &errorCode) 
{
	TRACE(dlSmModule,1,("Entering DlResetReq\n"));

	// now, go ahead and send a SABM

	_dl->XmitFrame(errorCode,TRUE,SABM,1);

	if ( errorCode == codeSuccessful ) {		
		// Start T1 timer, and change state.
		_dl->StartT1Timer(FALSE);					
		_dl->_setupType = externalReset;
		CHANGE_STATE(dlOutSetupPending);
	}
	else 
		return this;
}

/*
 * Function : HandleFrmrCondition()
 *
 * Description : This function is called in order to format
 * an FRMR response, and transmit it. 'type' give the
 *	the W,X,Y,Z type of the FRMR response.
 * _dl->_frmrInfo[] is an array of 3 unsigned chars that
 * holds the frmr info. field, and is available if the
 * same FRMR response needs to be retransmitted.
 * Change state to dlFrameReject after the FRMR has been
 * transmitted.
 */


dlStateMachine *
dlInfoXfer::HandleFrmrCondition(FrmrConditionType type, int &errorCode) 
{
	TRACE(dlSmModule,1,("Entering HandleFrmrCondtion. Type : %d\n",type));

	// Form the information field for the FRMR frame
	// First clear the bytes
	_dl->_frmrInfo[0] &= 0x00;
	_dl->_frmrInfo[1] &= 0x00;
	_dl->_frmrInfo[2] &= 0x00;

	// The first byte is the control byte of the received frame
	// which caused the rejection

	_dl->_frmrInfo[0] = _dl->_lastFrameRecd->GetCntl(); // Control byte  	

	// The second byte contains V(r), V(s) and whether the received
	// frame was a response or command frame.

	_dl->_frmrInfo[1] =
		(_dl->_win.GetVs() << 1) | 
		(( _dl->_lastFrameRecd->IsCmd() == TRUE ? 0 : 1 ) << 4 ) |
		( _dl->_win.GetVr() << 5 ); 

	// The third byte..
	_dl->_frmrInfo[2] |= type;
	if ( type == X )
		_dl->_frmrInfo[2] |= W; // set the 'W' bit also

	// Transmit the FRMR
	_dl->XmitFrmr(_dl->_lastFrameRecd->GetPf(),errorCode);
	if ( errorCode == codeSuccessful ) {		
		_dl->StartT1Timer(FALSE);					
		CHANGE_STATE(dlFrameReject);
	}
	else 
		return this;
}

/*
 * Function : HandleIframe()
 *
 * Description : Processes an I-frame that was received.
 * First the N(r) value is processed. If the N(r) is invalid
 * invoke HandleFrmrCondition(). If N(s) is out of sequence,
 * transmit a REJ, start rej timer and change state to dlReject.
 * If both N(r) and N(s) are allright, proceed. 
 */

dlStateMachine *
dlInfoXfer::HandleIframe(int &errorCode)
{
	_dl->_lastFrameRecd->PrintParams();

	// Check for ack
	// Transmit FRMR if invalid N(r)
	int nr = _dl->_lastFrameRecd->GetNr();
	int numAcked;
	if ( (numAcked = _dl->_win.ProcessAck(nr)) == -1 )
		return HandleFrmrCondition(Z,errorCode);

	unsigned int ns = _dl->_lastFrameRecd->GetNs();

	if ( _dl->_win.CheckSequence(ns) == FALSE ) {
		// Out of sequence I-frame
		// Transmit REJ, and go into dlReject state

		TRACE(dlSmModule,1,("Out of sequence frame\n"));

		_dl->XmitFrame(errorCode,FALSE,REJ,
				_dl->_lastFrameRecd->GetPf(),_dl->_win.GetVr());
		_dl->StartRejTimer(FALSE);
		CHANGE_STATE(dlReject);
	}

	if ( _dl->_lastFrameRecd->GetPf()) {
		// Have to send a response since P=1
		SendStateResponse();
	}
	else if ( !(_dl->_win.GetQueueLength() > _dl->_win.UnackedFrames())) {
		// There is at no I-frame waiting to be transmitted.
		// Send a acknowledgement now in a supervisory frame.
		SendStateResponse();
	}

	// Else, the acknowledgement will be sent in the next I-frame.		
	// Notify the network layer

	_dl->_nPlus1->Command(DL_DataInd, errorCode,
		_dl->_lastFrameRecd->GetDataPacket());

	// Now, if the receiverBusy flag is set to TRUE, we are to
	// go into a receiver busy state. remember this is just a 
 	// simulation of the receiver busy condition, since our 
 	// software cannot really go into a busy state.

	if ( receiverBusy == FALSE )
		return this;

	// Receiver is busy - change state
	CHANGE_STATE(dlReceiverBusy);
}

/*
 * Function : HandleRr()
 *
 * Description : Processes an RR that was received. First
 * process the N(r) field. If invalid, call HandleFrmrCondition().
 * The Checkpointing cycle if ongoing(evidenced by the a T1 timer
 * that is running) then stop the cycle only if a response with
 * F=1 is received. Once cycle is stopped, retransmit any old
 * and transmit any new I-frames if necessary.
 */

dlStateMachine *
dlInfoXfer::HandleRr(int &errorCode)
{
	TRACE(dlSmModule,1,("In HandleRr\n"));
	// Check for ack
	// Transmit FRMR if invalid N(r)
	int nr = _dl->_lastFrameRecd->GetNr();

	int numAcked;
	if ( (numAcked = _dl->_win.ProcessAck(nr)) == -1 )
		return HandleFrmrCondition(Z,errorCode);

	Boolean cmd = _dl->_lastFrameRecd->IsCmd();

	if ( _dl->_lastFrameRecd->GetPf() ) { // P/F = 1
		if ( cmd == TRUE )
			SendStateResponse();
		else if ( _dl->_remoteBusy == TRUE || 
							_dl->_t1Timer != 0 ) { 
			// Remote DTE was busy
			// A checkpointing cycle just finished. 
			// Retransmit any frames if necessary.

			TRACE(dlSmModule,2,("Recd. resp for receiver status command\n"));
			Transmit(TRUE); // re-transmit any frames
		}
	}

	_dl->_remoteBusy = FALSE; // Remote DTE is no longer busy
	return this;
}
 
/*
 * Function : HandleRnr()
 *
 * Description : Processes a RNR that was received. First
 * process N(r). If invalid, invoke HandleFrmrCondition().
 * Start Timer T1 afresh, set _remoteBusy flag to TRUE. 
 * No more I-frames can be transmitted unless the remote DTE
 * busy condition is cleared by a RR or a REJ.
 */

dlStateMachine *
dlInfoXfer::HandleRnr(int &errorCode)
{
	TRACE(dlSmModule,1,("In HandleRnr\n"));

	// Check for ack
	// Transmit FRMR if invalid N(r)
	int nr = _dl->_lastFrameRecd->GetNr();
	int numAcked;
	if ( (numAcked = _dl->_win.ProcessAck(nr)) == -1 )
		return HandleFrmrCondition(Z,errorCode);

	Boolean cmd = _dl->_lastFrameRecd->IsCmd();

	// Transmit a response immediately RNR is a command
	// with P = 1
	if ( _dl->_lastFrameRecd->GetPf() && cmd == TRUE )
		SendStateResponse();


	// Transmit a supervisory command querying for the 
	// remote DTE status.

	_dl->XmitFrame(errorCode,TRUE,GetSupervisoryFrame(),1,
				_dl->_win.GetVr());
	_dl->StartT1Timer(FALSE); // Start timer T1 with a zero retry count
	_dl->_remoteBusy = TRUE; // Remote DTE is busy now
	return this;
}
 
/*
 * Function : HandleRej()
 *
 * Description : Processes a REJ that was received.
 * First check N(r), if invalid call HandleFrmrCondition().
 * Reset any remote DTE busy condition, and re-transmit
 * frames that have not been acknowledged.
 */

dlStateMachine *
dlInfoXfer::HandleRej(int &errorCode)
{
	// Check for ack
	// Transmit FRMR if invalid N(r)

	TRACE(dlSmModule,1,("In HandleRej\n"));

	int nr = _dl->_lastFrameRecd->GetNr();
	int numAcked;
	if ( (numAcked = _dl->_win.ProcessAck(nr)) == -1 )
		return HandleFrmrCondition(Z,errorCode);

	Boolean cmd = _dl->_lastFrameRecd->IsCmd();
	
	if ( cmd == TRUE && _dl->_lastFrameRecd->GetPf() )
			SendStateResponse();

	// Clear any remote DTE busy condition 
	_dl->_remoteBusy = FALSE; 

	// Start retransmissions if needed 
	// Note Transmit() is called with argument TRUE -
	// This causes the window to be reset for retransmission. 

	Transmit(TRUE); 
	return this;
}
 
/*
 * Function : HandleT1Timer()
 *
 * Description : Handles a T1 timer expiry event.
 * 	A supervisory command needs to be retransmitted.
 *	Since the same function is used in dlInfoXfer,
 *	dlReject and dlReceiverBusy state, the proper
 *	supervisory command is got by invoking 
 *	GetSupervisoryFrame(). This is defined in dlInfoXfer
 *	to return RR, and redefined in dlReceiverBusy and
 * 	dlReject to return RNR and REJ respectively.
 *	If supervisory command has been retransmitted upto
 *	N2 times, start link reset.
 */

dlStateMachine *
dlInfoXfer::HandleT1Timer(int &errorCode)
{
	TRACE(dlSmModule,1,("In HandleT1Timer\n"));

	if ( _dl->_t1RetryCount < N2 ) {

		// Transmit supervisory command frame
		// GetSupervisoryFrame() returns the type of 
		// supervisory command to send.

		_dl->XmitFrame(errorCode,TRUE,GetSupervisoryFrame(),1,
				_dl->_win.GetVr());

		// Restart T1
		_dl->StartT1Timer(TRUE);
		return this;
	}
	else { 
		// Already re-transmitted N2 times. Reset.
		TRACE(dlSmModule,1,("Timer T1 timed out N2 times\n"));
		return (LinkReset(errorCode));
	}
}


/*
 * Function : CheckT1Status()
 *
 * Description : Checks the status of the T1 timer. If
 *	any I-frames are still unacknowledged, the timer
 *	is restarted (with retry count 0). Else, the
 *	timer is stopped.
 */

void	
dlInfoXfer::CheckT1Status()
{
	if ( _dl->_win.UnackedFrames() != 0 ) {

		// There are unacknowledged I-frames.
		// Start T1 timer again.

		TRACE(dlSmModule,1,("Restarting timer T1\n"));
		_dl->StartT1Timer(FALSE);
		_dl->_expFBit = 1;
	}
	else {
		TRACE(dlSmModule,1,("Stopping Timer T1\n"));
		_dl->StopT1Timer();
	}
}

/*
 * Function : Transmit()
 *
 * Description : This function transmits any I-frames
 *	outstanding (as many as the window allows).
 *  If this is a retransmission, then the window is
 *	first reset. 
 */

void
dlInfoXfer::Transmit(Boolean retransmit)
{
	int errorCode = codeSuccessful;

	if ( retransmit == TRUE )
		_dl->_win.StartRetransmission(); // reset V(s)

	// Query the window for I-frames to transmit
	// The window will return I-frames with the
	// proper N(s), N(r) etc.

	dataPacket *dp;
	while ( dp = _dl->_win.GetNextToXmit() ) {
		_dl->_nMinus1->Command(PH_DataReq,errorCode,dp);
		if ( errorCode != codeSuccessful )
			return;
		_dl->_expFBit = 1; // Because the last I-frame transmitted with P=1
	} 
	CheckT1Status(); // Start timer T1 if necessary
}

/*
 * Function : SendStateResponse()
 *
 * Description : Transmits a supervisory response. The
 * current state of the DTE is indicated by the
 * return value of the GetSupervisoryFrame() which
 * returns RR in dlInfoXfer, and is redefined to return
 * RNR and REJ in the dlReceiverBusy and dlReject states
 * respectively.
 */

void
dlInfoXfer::SendStateResponse()
{
	// Transmit a supervisory response.
	int errorCode = codeSuccessful;
	_dl->XmitFrame(errorCode,FALSE,GetSupervisoryFrame(),
		_dl->_lastFrameRecd->GetPf(), _dl->_win.GetVr());

	if ( errorCode != codeSuccessful)
		TRACE(dlSmModule,1,("Error sending supervisory response\n"));
}

/*
 * Function : GetSupervisoryFrame()
 *
 * Description : Returns current receiver status.
 */

int 
dlInfoXfer::GetSupervisoryFrame()
{
	return RR;
}

/*
 * Function : HandleUndefined()
 *
 * Description : Processes a frame with undefined control
 *	field. Invokes HandleFrmrCondition, with W bit set.
 */

dlStateMachine *
dlInfoXfer::HandleUndefined(int &errorCode)
{
	TRACE(dlSmModule,1,("Handling undefined FRAME\n"));
	return HandleFrmrCondition(W,errorCode);
}

/*
 * Function : HandleIncorrectLength()
 *
 * Description : Processes a frame with incorrect length.
 *	Invokes HandleFrmrCondition, with X bit set.
 */


dlStateMachine *
dlInfoXfer::HandleIncorrectLength(int &errorCode)
{
	TRACE(dlSmModule,1,("Handling FRAME with incorrect length\n"));
	return HandleFrmrCondition(X,errorCode);
}

/*
 * Function : HandleIfrLengthExcess()
 *
 * Description : Processes a I-frame with excess info. 
 *	field length. Invokes HandleFrmrCondition, with Y bit set.
 */


dlStateMachine *
dlInfoXfer::HandleIfrLengthExcess(int &errorCode)
{
	TRACE(dlSmModule,1,("Handling IFRAME with excess length\n"));
	return HandleFrmrCondition(Y,errorCode);
}



