/*
 * File : dlOutSetupPending.C
 *
 * Description : This file contains the method definitions for
 * the dlOutSetupPending class.
 */

#include "phEvents.h"
#include "dlEvents.h"
#include "dllayer.h"
#include "dlDisconnected.h"
#include "dlOutSetupPending.h"
#include "dlDiscPending.h"
#include "dlInfoXfer.h"
#include "frame.h"

/*
 * Function : PrintState()
 * 
 * Description : Prints the state.
 */
 
void 
dlOutSetupPending::PrintState() 
{
	TRACE(dlSmModule,1,("State dlOutSetupPending\n"));
}

/*
 * Function : HandleT1Timer()
 * 
 * Description : Process a timer T1 timeout event. Re-transmit
 * a SABM. If SABM has already been retransmitted N2 times,
 * invoke DL_DisconnectInd, and change state to dlDisconnected.
 */
 
dlStateMachine *
dlOutSetupPending::HandleT1Timer(int &errorCode)
{
	TRACE(dlSmModule,1,("dlOutSetupPending: In HandleT1Timer\n"));
	if ( _dl->_t1RetryCount < N2 ) {

		// Transmit a SABM
		_dl->XmitFrame(errorCode,TRUE,SABM,1);
		if ( errorCode == codeSuccessful ) {		
			_dl->StartT1Timer(TRUE); // re-start T1 again					
			return this;
		}
		TRACE(dlSmModule,1,("Failed to transmit SABM!\n"));
		_dl->_nPlus1->Command(DL_DisconnectInd,errorCode,noPeerResponse);
		CHANGE_STATE(dlDisconnected);
	}
	else { 
		TRACE(dlSmModule,1,("Timer T1 timed out N2 times\n"));
		_dl->_nPlus1->Command(DL_DisconnectInd,errorCode,noPeerResponse);
		CHANGE_STATE(dlDisconnected);
	}
}

/*
 * Function : HandleUa()
 * 
 * Description : Processes a UA that has been received.
 *	This is response to a SABM. Depending on the setup type,
 * 	invoke the appropriate primitive on the network layer,
 *	notifying that the link is setup/ reset. Change
 *	state to dlInfoXfer.
 */
 
dlStateMachine *
dlOutSetupPending::HandleUa(int &errorCode)
{
	if ( !_dl->_lastFrameRecd->GetPf())
		return this; // Ignore UA with F=0, we are expecting F=1

	TRACE(dlSmModule,1,("In HandleUA\n"));
	if ( _dl->_setupType == initial )
		_dl->_nPlus1->Command(DL_ConnectCnf,errorCode);
	else if ( _dl->_setupType == externalReset )
		_dl->_nPlus1->Command(DL_ResetCnf,errorCode);
	else 
		_dl->_nPlus1->Command(DL_ResetInd,errorCode,internalError);
	CHANGE_TO_INFOXFER(TRUE);
}

/*
 * Function : HandleDm()
 * 
 * Description : Processes a DM that was received. 
 * This is in response to a SABM. Change state
 * to dlDisconnected, and notify network layer of
 * link disconnection. 
 */
 
dlStateMachine *
dlOutSetupPending::HandleDm(int &errorCode)
{
	// Ignore a mode setting command, i.e., DM F=0
	if ( !_dl->_lastFrameRecd->GetPf())
		return this;

	_dl->_nPlus1->Command(DL_DisconnectInd,errorCode,peerDisconnect);
	CHANGE_STATE(dlDisconnected);	
}
 
/*
 * Function : HandleDisc()
 * 
 * Description : Processes a DISC that was received.
 *	This is a collision of dissimilar unnumbered commands.
 *	Transmit a DM, change state to dlDisconnected and
 * 	notify network layer.
 */
 
dlStateMachine *
dlOutSetupPending::HandleDisc(int &errorCode)
{
	_dl->XmitFrame(errorCode,FALSE,DM,_dl->_lastFrameRecd->GetPf());
	_dl->_nPlus1->Command(DL_DisconnectInd,errorCode,peerDisconnect);
	CHANGE_STATE(dlDisconnected);
}
 
/*
 * Function : HandleSabm()
 * 
 * Description : Processes a SABM that was received.
 *	This is a collision of similar unnumbered commands.
 *	Transmit a UA, notify network layer of link setup/reset,
 *	and change state to dlInfoXfer.
 */
 
dlStateMachine *
dlOutSetupPending::HandleSabm(int &errorCode)
{
	// Transmit a UA
	_dl->XmitFrame(errorCode,FALSE,UA,_dl->_lastFrameRecd->GetPf());
	if ( _dl->_setupType == initial ) 
		_dl->_nPlus1->Command(DL_ConnectCnf,errorCode);
	else if ( _dl->_setupType == externalReset )
		_dl->_nPlus1->Command(DL_ResetCnf,errorCode);
	else 
		_dl->_nPlus1->Command(DL_ResetInd,errorCode,internalError);
	CHANGE_TO_INFOXFER(TRUE);
}

/*
 * Function : DlDisconnectReq()
 * 
 * Description : Processes a DL_DisconnectReq primitive invoked
 *	by the network layer. Transmit a DISC, change state
 *	to dlDiscPending.
 */
 
dlStateMachine * 
dlOutSetupPending::DlDisconnectReq(int &errorCode)
{
	_dl->XmitFrame(errorCode,TRUE,DISC,1);
	CHANGE_STATE(dlDiscPending);
}

