/*
 * File : trace.h
 *
 * Description : This file contains classes trace and
 * traceManager, definitions for macros 
 * DEFINE_TRACE_MODULE, DECLARE_TRACE_MODULE and TRACE.
 */


#ifndef _trace_h_
#define _trace_h_

#include <stdio.h>
#include "types.h"

const int MaxModules = 64;

/*
 * Class : trace
 *
 * Description : Each instance of this class represents
 * the tracing status of one module. Tracing levels can
 * be set, reset. Tracing can be turned off/on, and trace
 * information can be printed.
 *
 * Attributes:
 *	int _level : The trace level for this module.
 *	Boolean _on : If True - trace is on for this module.
 *	char *_name : The name of this module.
 */

class trace {
 public:
	trace(const char *name);
	~trace(void);
	Boolean IsModuleOn(void);
	int	Level(void);
	const char *Name(void);
	void SetTrace(int level);
	void ResetTrace(void);
	static void PrintTraceInfo(const char *,...);
	void PrintModuleName(void);
 private:
	int	_level;
	Boolean _on;
	char * _name;
};
	
/*
 * Class : traceManager
 *
 * Description : This class represents a collection of
 * trace objects. Provides functions to print info about
 * all the trace objects, get the address of a given 
 * trace object, and add new trace objects.
 *
 * Attributes :
 *	int _index : The index for the next module to be added 
 *			the list of modules.
 *	trace *_modules[] : An array of pointers to trace objects.
 */

class traceManager {
 public:
	static void AddModule(trace *module);
	static void PrintModules();
	static trace *GetModule(int index);
 private:
	static int _index;
	static trace *_modules[MaxModules];		
};

/*
 * Macro : TRACE
 *
 * Description : Prints out the trace information for
 * a given module, if the tracing is on, and the level
 * is greater than the tracing level of the trace object
 * for that module. Note that defining NTRACE during
 * compilation will result in a NULL expansion of the
 * macro.
 */

#ifdef NTRACE 
#define TRACE(module,level,args)
#else

#define TRACE(module,level,args) 		\
	if (module.IsModuleOn() == TRUE && 	\
				level >= module.Level()) 	\
	{ 												\
		module.PrintModuleName(); 			\
		(module.PrintTraceInfo args); 	\
	}
#endif

/*
 * Macro : DECLARE_TRACE_MODULE
 *
 * Description : Provides an extern declaration of a 
 * given trace object.
 */


#define DECLARE_TRACE_MODULE(module) \
	extern trace module;

/*
 * Macro : DEFINE_TRACE_MODULE
 *
 * Description : Instantiates a trace object for the
 * given module.
 */

#define DEFINE_TRACE_MODULE(module,name) \
	trace module(name);
	

#endif // _trace_h_
